@@ -26,18 +26,21 @@ from botocore.model import ServiceModel
 from botocore.paginate import Paginator
 from botocore.utils import CachedProperty
 from botocore.utils import get_service_module_name
-from botocore.utils import switch_host_s3_accelerate
 from botocore.utils import S3RegionRedirector
-from botocore.utils import fix_s3_host
-from botocore.utils import switch_to_virtual_host_style
-from botocore.utils import S3_ACCELERATE_WHITELIST
+from botocore.utils import S3ArnParamHandler
+from botocore.utils import S3EndpointSetter
 from botocore.args import ClientArgsCreator
-from botocore.compat import urlsplit
 from botocore import UNSIGNED
 # Keep this imported.  There's pre-existing code that uses
 # "from botocore.client import Config".
 from botocore.config import Config
 from botocore.history import get_global_history_recorder
+from botocore.discovery import (
+    EndpointDiscoveryHandler, EndpointDiscoveryManager,
+    block_endpoint_discovery_required_operations
+)
+from botocore.retries import standard
+from botocore.retries import adaptive
 
 
 logger = logging.getLogger(__name__)
@@ -48,7 +51,8 @@ class ClientCreator(object):
     """Creates client objects for a service."""
     def __init__(self, loader, endpoint_resolver, user_agent, event_emitter,
                  retry_handler_factory, retry_config_translator,
-                 response_parser_factory=None, exceptions_factory=None):
+                 response_parser_factory=None, exceptions_factory=None,
+                 config_store=None):
         self._loader = loader
         self._endpoint_resolver = endpoint_resolver
         self._user_agent = user_agent
@@ -57,6 +61,11 @@ class ClientCreator(object):
         self._retry_config_translator = retry_config_translator
         self._response_parser_factory = response_parser_factory
         self._exceptions_factory = exceptions_factory
+        # TODO: Migrate things away from scoped_config in favor of the
+        # config_store.  The config store can pull things from both the scoped
+        # config and environment variables (and potentially more in the
+        # future).
+        self._config_store = config_store
 
     def create_client(self, service_name, region_name, is_secure=True,
                       endpoint_url=None, verify=None,
@@ -79,6 +88,9 @@ class ClientCreator(object):
         self._register_s3_events(
             service_client, endpoint_bridge, endpoint_url, client_config,
             scoped_config)
+        self._register_endpoint_discovery(
+            service_client, endpoint_url, client_config
+        )
         return service_client
 
     def create_client_class(self, service_name, api_version=None):
@@ -106,6 +118,26 @@ class ClientCreator(object):
         return service_model
 
     def _register_retries(self, client):
+        retry_mode = client.meta.config.retries['mode']
+        if retry_mode == 'standard':
+            self._register_v2_standard_retries(client)
+        elif retry_mode == 'adaptive':
+            self._register_v2_standard_retries(client)
+            self._register_v2_adaptive_retries(client)
+        elif retry_mode == 'legacy':
+            self._register_legacy_retries(client)
+
+    def _register_v2_standard_retries(self, client):
+        max_attempts = client.meta.config.retries.get('total_max_attempts')
+        kwargs = {'client': client}
+        if max_attempts is not None:
+            kwargs['max_attempts'] = max_attempts
+        standard.register_retry_handler(**kwargs)
+
+    def _register_v2_adaptive_retries(self, client):
+        adaptive.register_retry_handler(client)
+
+    def _register_legacy_retries(self, client):
         endpoint_prefix = client.meta.service_model.endpoint_prefix
         service_id = client.meta.service_model.service_id
         service_event_name = service_id.hyphenize()
@@ -116,10 +148,11 @@ class ClientCreator(object):
         if not original_config:
             return
 
+        retries = self._transform_legacy_retries(client.meta.config.retries)
         retry_config = self._retry_config_translator.build_retry_config(
             endpoint_prefix, original_config.get('retry', {}),
             original_config.get('definitions', {}),
-            client.meta.config.retries
+            retries
         )
 
         logger.debug("Registering retry handlers for service: %s",
@@ -132,103 +165,62 @@ class ClientCreator(object):
             unique_id=unique_id
         )
 
+    def _transform_legacy_retries(self, retries):
+        if retries is None:
+            return
+        copied_args = retries.copy()
+        if 'total_max_attempts' in retries:
+            copied_args = retries.copy()
+            copied_args['max_attempts'] = (
+                copied_args.pop('total_max_attempts') - 1)
+        return copied_args
+
+    def _get_retry_mode(self, client, config_store):
+        client_retries = client.meta.config.retries
+        if client_retries is not None and \
+                client_retries.get('mode') is not None:
+            return client_retries['mode']
+        return config_store.get_config_variable('retry_mode') or 'legacy'
+
+    def _register_endpoint_discovery(self, client, endpoint_url, config):
+        if endpoint_url is not None:
+            # Don't register any handlers in the case of a custom endpoint url
+            return
+        # Only attach handlers if the service supports discovery
+        if client.meta.service_model.endpoint_discovery_operation is None:
+            return
+        events = client.meta.events
+        service_id = client.meta.service_model.service_id.hyphenize()
+        enabled = False
+        if config and config.endpoint_discovery_enabled is not None:
+            enabled = config.endpoint_discovery_enabled
+        elif self._config_store:
+            enabled = self._config_store.get_config_variable(
+                'endpoint_discovery_enabled')
+        if enabled:
+            manager = EndpointDiscoveryManager(client)
+            handler = EndpointDiscoveryHandler(manager)
+            handler.register(events, service_id)
+        else:
+            events.register('before-parameter-build',
+                            block_endpoint_discovery_required_operations)
+
     def _register_s3_events(self, client, endpoint_bridge, endpoint_url,
                             client_config, scoped_config):
         if client.meta.service_model.service_name != 's3':
             return
         S3RegionRedirector(endpoint_bridge, client).register()
-        self._set_s3_addressing_style(
-            endpoint_url, client.meta.config.s3, client.meta.events,
-            client.meta.partition
-        )
-        # Enable accelerate if the configuration is set to to true or the
-        # endpoint being used matches one of the accelerate endpoints.
-        if self._is_s3_accelerate(endpoint_url, client.meta.config.s3):
-            # Also make sure that the hostname gets switched to
-            # s3-accelerate.amazonaws.com
-            client.meta.events.register_first(
-                'before-sign.s3', switch_host_s3_accelerate)
-
+        S3ArnParamHandler().register(client.meta.events)
+        S3EndpointSetter(
+            endpoint_resolver=self._endpoint_resolver,
+            region=client.meta.region_name,
+            s3_config=client.meta.config.s3,
+            endpoint_url=endpoint_url,
+            partition=client.meta.partition
+        ).register(client.meta.events)
         self._set_s3_presign_signature_version(
             client.meta, client_config, scoped_config)
 
-    def _set_s3_addressing_style(self, endpoint_url, s3_config, event_emitter,
-                                 partition):
-        if s3_config is None:
-            s3_config = {}
-
-        addressing_style = self._get_s3_addressing_style(
-            endpoint_url, s3_config)
-        handler = self._get_s3_addressing_handler(
-            endpoint_url, s3_config, addressing_style, partition)
-        if handler is not None:
-            event_emitter.register('before-sign.s3', handler)
-
-    def _get_s3_addressing_style(self, endpoint_url, s3_config):
-        # Use virtual host style addressing if accelerate is enabled or if
-        # the given endpoint url is an accelerate endpoint.
-        accelerate = s3_config.get('use_accelerate_endpoint', False)
-        if accelerate or self._is_s3_accelerate(endpoint_url, s3_config):
-            return 'virtual'
-
-        # If a particular addressing style is configured, use it.
-        configured_addressing_style = s3_config.get('addressing_style')
-        if configured_addressing_style:
-            return configured_addressing_style
-
-    def _get_s3_addressing_handler(self, endpoint_url, s3_config,
-                                   addressing_style, partition):
-        # If virtual host style was configured, use it regardless of whether
-        # or not the bucket looks dns compatible.
-        if addressing_style == 'virtual':
-            logger.debug("Using S3 virtual host style addressing.")
-            return switch_to_virtual_host_style
-
-        # If path style is configured, no additional steps are needed. If
-        # endpoint_url was specified, don't default to virtual. We could
-        # potentially default provided endpoint urls to virtual hosted
-        # style, but for now it is avoided.
-        if addressing_style == 'path' or endpoint_url is not None:
-            logger.debug("Using S3 path style addressing.")
-            return None
-
-        logger.debug("Defaulting to S3 virtual host style addressing with "
-                     "path style addressing fallback.")
-
-        # By default, try to use virtual style with path fallback.
-        return fix_s3_host
-
-    def _is_s3_accelerate(self, endpoint_url, s3_config):
-        # Accelerate has been explicitly configured.
-        if s3_config is not None and s3_config.get('use_accelerate_endpoint'):
-            return True
-
-        # Accelerate mode is turned on automatically if an endpoint url is
-        # provided that matches the accelerate scheme.
-        if endpoint_url is None:
-            return False
-
-        # Accelerate is only valid for Amazon endpoints.
-        netloc = urlsplit(endpoint_url).netloc
-        if not netloc.endswith('amazonaws.com'):
-            return False
-
-        # The first part of the url should always be s3-accelerate.
-        parts = netloc.split('.')
-        if parts[0] != 's3-accelerate':
-            return False
-
-        # Url parts between 's3-accelerate' and 'amazonaws.com' which
-        # represent different url features.
-        feature_parts = parts[1:-2]
-
-        # There should be no duplicate url parts.
-        if len(feature_parts) != len(set(feature_parts)):
-            return False
-
-        # Remaining parts must all be in the whitelist.
-        return all(p in S3_ACCELERATE_WHITELIST for p in feature_parts)
-
     def _set_s3_presign_signature_version(self, client_meta,
                                           client_config, scoped_config):
         # This will return the manually configured signature version, or None
@@ -242,9 +234,13 @@ class ClientCreator(object):
         # Check to see if the region is a region that we know about. If we
         # don't know about a region, then we can safely assume it's a new
         # region that is sigv4 only, since all new S3 regions only allow sigv4.
+        # The only exception is aws-global. This is a pseudo-region for the
+        # global endpoint, we should respect the signature versions it
+        # supports, which includes v2.
         regions = self._endpoint_resolver.get_available_endpoints(
             's3', client_meta.partition)
-        if client_meta.region_name not in regions:
+        if client_meta.region_name != 'aws-global' and \
+                client_meta.region_name not in regions:
             return
 
         # If it is a region we know about, we want to default to sigv2, so here
@@ -285,7 +281,7 @@ class ClientCreator(object):
         args_creator = ClientArgsCreator(
             self._event_emitter, self._user_agent,
             self._response_parser_factory, self._loader,
-            self._exceptions_factory)
+            self._exceptions_factory, config_store=self._config_store)
         return args_creator.get_client_args(
             service_model, region_name, is_secure, endpoint_url,
             verify, credentials, scoped_config, client_config, endpoint_bridge)
@@ -379,13 +375,14 @@ class ClientEndpointBridge(object):
 
     def _create_endpoint(self, resolved, service_name, region_name,
                          endpoint_url, is_secure):
+        explicit_region = region_name is not None
         region_name, signing_region = self._pick_region_values(
             resolved, region_name, endpoint_url)
         if endpoint_url is None:
             if self._is_s3_dualstack_mode(service_name):
                 endpoint_url = self._create_dualstack_endpoint(
                     service_name, region_name,
-                    resolved['dnsSuffix'], is_secure)
+                    resolved['dnsSuffix'], is_secure, explicit_region)
             else:
                 # Use the sslCommonName over the hostname for Python 2.6 compat.
                 hostname = resolved.get('sslCommonName', resolved.get('hostname'))
@@ -421,7 +418,12 @@ class ClientEndpointBridge(object):
         return False
 
     def _create_dualstack_endpoint(self, service_name, region_name,
-                                   dns_suffix, is_secure):
+                                   dns_suffix, is_secure, explicit_region):
+        if not explicit_region and region_name == 'aws-global':
+            # If the region_name passed was not explicitly set, default to
+            # us-east-1 instead of the modeled default aws-global. Dualstack
+            # does not support aws-global
+            region_name = 'us-east-1'
         hostname = '{service}.dualstack.{region}.{dns_suffix}'.format(
             service=service_name, region=region_name,
             dns_suffix=dns_suffix)
@@ -607,8 +609,8 @@ class BaseClient(object):
         if event_response is not None:
             http, parsed_response = event_response
         else:
-            http, parsed_response = self._endpoint.make_request(
-                operation_model, request_dict)
+            http, parsed_response = self._make_request(
+                operation_model, request_dict, request_context)
 
         self.meta.events.emit(
             'after-call.{service_id}.{operation_name}'.format(
@@ -625,6 +627,18 @@ class BaseClient(object):
         else:
             return parsed_response
 
+    def _make_request(self, operation_model, request_dict, request_context):
+        try:
+            return self._endpoint.make_request(operation_model, request_dict)
+        except Exception as e:
+            self.meta.events.emit(
+                'after-call-error.{service_id}.{operation_name}'.format(
+                    service_id=self._service_model.service_id.hyphenize(),
+                    operation_name=operation_model.name),
+                exception=e, context=request_context
+            )
+            raise
+
     def _convert_to_request_dict(self, api_params, operation_model,
                                  context=None):
         api_params = self._emit_api_params(
