@@ -2,6 +2,7 @@ from tests import unittest
 
 from botocore import model
 from botocore.compat import OrderedDict
+from botocore.exceptions import MissingServiceIdError
 
 
 def test_missing_model_attribute_raises_exception():
@@ -30,6 +31,16 @@ def test_missing_model_attribute_raises_exception():
         yield _test_attribute_raise_exception, name
 
 
+class TestServiceId(unittest.TestCase):
+    def test_hypenize_replaces_spaces(self):
+        self.assertEqual(
+            model.ServiceId('my service').hyphenize(), 'my-service'
+        )
+
+    def test_hyphenize_lower_cases(self):
+        self.assertEqual(model.ServiceId('MyService').hyphenize(), 'myservice')
+
+
 class TestServiceModel(unittest.TestCase):
 
     def setUp(self):
@@ -61,6 +72,28 @@ class TestServiceModel(unittest.TestCase):
     def test_service_id(self):
         self.assertEqual(self.service_model.service_id, 'MyService')
 
+    def test_hyphenize_service_id(self):
+        self.assertEqual(
+            self.service_model.service_id.hyphenize(), 'myservice')
+
+    def test_service_id_does_not_exist(self):
+        service_model = {
+            'metadata': {
+                'protocol': 'query',
+                'endpointPrefix': 'endpoint-prefix',
+            },
+            'documentation': 'Documentation value',
+            'operations': {},
+            'shapes': {
+                'StringShape': {'type': 'string'}
+            }
+        }
+        service_name = 'myservice'
+        service_model = model.ServiceModel(service_model, service_name)
+        with self.assertRaisesRegexp(model.UndefinedModelAttributeError,
+                                     service_name):
+            service_model.service_id
+
     def test_operation_does_not_exist(self):
         with self.assertRaises(model.OperationNotFoundError):
             self.service_model.operation_model('NoExistOperation')
@@ -75,6 +108,10 @@ class TestServiceModel(unittest.TestCase):
     def test_shape_names(self):
         self.assertEqual(self.service_model.shape_names, ['StringShape'])
 
+    def test_repr_has_service_name(self):
+        self.assertEqual(repr(self.service_model),
+                         'ServiceModel(endpoint-prefix)')
+
 
 class TestOperationModelFromService(unittest.TestCase):
     def setUp(self):
@@ -246,6 +283,33 @@ class TestOperationModelFromService(unittest.TestCase):
         operation_two = service_model.operation_model('OperationTwo')
         self.assertFalse(operation_two.deprecated)
 
+    def test_endpoint_operation_present(self):
+        self.model['operations']['OperationName']['endpointoperation'] = True
+        service_model = model.ServiceModel(self.model)
+        operation_name = service_model.operation_model('OperationName')
+        self.assertTrue(operation_name.is_endpoint_discovery_operation)
+
+    def test_endpoint_operation_present_false(self):
+        self.model['operations']['OperationName']['endpointoperation'] = False
+        service_model = model.ServiceModel(self.model)
+        operation_name = service_model.operation_model('OperationName')
+        self.assertFalse(operation_name.is_endpoint_discovery_operation)
+
+    def test_endpoint_operation_absent(self):
+        operation_two = self.service_model.operation_model('OperationName')
+        self.assertFalse(operation_two.is_endpoint_discovery_operation)
+
+    def test_endpoint_discovery_present(self):
+        operation = self.model['operations']['OperationName']
+        operation['endpointdiscovery'] = {'required': True}
+        service_model = model.ServiceModel(self.model)
+        operation_name = service_model.operation_model('OperationName')
+        self.assertTrue(operation_name.endpoint_discovery.get('required'))
+
+    def test_endpoint_discovery_absent(self):
+        operation_name = self.service_model.operation_model('OperationName')
+        self.assertIsNone(operation_name.endpoint_discovery)
+
 
 class TestOperationModelEventStreamTypes(unittest.TestCase):
     def setUp(self):
@@ -638,6 +702,26 @@ class TestShapeResolver(unittest.TestCase):
         self.assertEqual(member.metadata['max'], 128)
         self.assertEqual(member.metadata['sensitive'], True)
 
+    def test_error_shape_metadata(self):
+        shapes = {
+            'ResourceNotFoundException': {
+                'type': 'structure',
+                'members': {
+                    'message': {
+                        'shape': 'ErrorMessage',
+                    }
+                },
+                'exception': True,
+                'retryable': {'throttling': True}
+            }
+        }
+        resolver = model.ShapeResolver(shapes)
+        shape = resolver.get_shape_by_name('ResourceNotFoundException')
+        self.assertEqual(
+            shape.metadata,
+            {'exception': True, 'retryable': {'throttling': True}}
+        )
+
     def test_shape_list(self):
         shapes = {
             'mfaDeviceListType': {
@@ -806,6 +890,20 @@ class TestBuilders(unittest.TestCase):
         self.assertEqual(shape.members['A'].documentation,
                          'MyDocs')
 
+    def test_min_max_used_in_metadata(self):
+        b = model.DenormalizedStructureBuilder()
+        shape = b.with_members({
+            'A': {
+                'type': 'string',
+                'documentation': 'MyDocs',
+                'min': 2,
+                'max': 3,
+            },
+        }).build_model()
+        metadata = shape.members['A'].metadata
+        self.assertEqual(metadata.get('min'), 2)
+        self.assertEqual(metadata.get('max'), 3)
+
     def test_use_shape_name_when_provided(self):
         b = model.DenormalizedStructureBuilder()
         shape = b.with_members({
